# -*- coding: utf-8 -*-
import shutil
import librosa
import numpy as np
from pathlib import Path
from scipy.io import wavfile
from multiprocessing import Pool
from functools import partial
from tqdm import tqdm, trange
from .tools import warn
from .AudioTools import check_input, load_waves
from .LoadConfig import CoreConfig

__all__ = ['Synthesizer']


class Synthesizer:
    def __repr__(self):
        return """
        syx = Synthesizer(clean, noise, out_path='', snr_ratio=(-5, 0, 5))
            clean, noise: 音頻資料, 可以是 
                  clean = 'data/clean' # str
                  clean = ['data/clean', 'data/clean2'] # list
                  clean = 'wave.txt' # .txt file
                  clean = ['wave.txt', 'data/clean2'] # list
            out_path: 資料輸出路徑

            snr_ratio: 混音 SNR 值, 可以是任何整數值, 一個或多個 
                  snr_ratio = [10]
                  snr_ratio = [-5, 0, 5]

        syx.syn(mode='regular', add_clean=False, rdn_choice=None)
            mode: 混音方法
                  mode = 'regular' 或 'increment'
                      mode: 'reg' or 'regular'
                          -> noiseWaves 按照 snrRatio & noise順序，循環加入cleanWaves(產生音檔數量與cleanWaves數量一樣)
                      mode: 'inc' or 'increment'
                          -> 產生音檔數量為 cleanWaves * noiseWaves * snrRatio
            add_clean: 是否將原始音檔輸出
            rdn_choice: 是否從輸入資料clean中隨機選取資料做為新的輸入資料clean

        syx.syn_backGround(background_path, background_snr=(5,), addClean, rdnChoice)
            將 noise 與 background_path的音檔 按照 background_snr 混合
            再 按照 snr_ratio 混入clean
            產生音檔數量為: clean*((1+noise)*(1+backgroundnoise)-1)*snr*(1+20%(silence))
            background_path: 輸入資料
            background_snr: 混音 SNR 值, 可以是任何整數值, 一個或多個
                  snr_ratio = [10]
                  snr_ratio = [-5, 0, 5]
            add_clean: 是否將原始音檔輸出
            rdn_choice: 是否從輸入資料clean中隨機選取資料做為新的輸入資料clean

        syn_waves(synWavesList, isSilence: bool, num=None)
            synWavesList = [[save_dir, clean_file, noise_file, snr], ... ]
        """

    def __init__(self, clean, noise, out_path='', snr_ratio=(-5, 0, 5)):
        # super(Synthesizer, self).__init__()
        # self.cleanPath, self.noisePath = Path(clean), Path(noise)
        assert all([Path(clean).is_dir(), Path(noise).is_dir()]), "input path error"

        self.cleanWaves, w = check_input(clean)
        self.noiseWaves, w = check_input(noise)
        assert self.cleanWaves, f"Error input {clean}, found {len(self.cleanWaves)} files."
        assert self.noiseWaves, f"Error input {noise}, found {len(self.noiseWaves)} files."

        self.noisyPath = Path('mixed' if out_path == '' else out_path)

        assert not isinstance(snr_ratio, str), f"type error: snr_ratio can't be str"
        self.snrRatio = snr_ratio if hasattr(snr_ratio, '__iter__') else [snr_ratio]

        self.synWavesList = None  # for syn processing

    def wave_input(self):
        return self.cleanWaves, self.noiseWaves

    def wave_syn(self) -> list:
        return self.synWavesList

    def waveform(self, is_pool=False):
        fcn = partial(load_waves, is_pool=is_pool)
        return fcn(self.cleanWaves), fcn(self.noiseWaves)

    def syn(self, mode='regular', add_clean=False, rdn_choice=None,
            is_overwrite=True, is_pool=True, is_silence=False, p_silence=0.02):
        print("[Mixed] ==== Synthesize Audio Processing ====")
        clean_waves = [np.random.choice(self.cleanWaves) for _ in range(rdn_choice)] if rdn_choice else self.cleanWaves
        clean_waves_len = len(clean_waves)
        noise_waves_len = len(self.noiseWaves)
        snrLen = len(self.snrRatio)
        print(f"Clean: {clean_waves_len}, Noise: {noise_waves_len}, SNR: {self.snrRatio[0]} ~ {self.snrRatio[-1]}")

        if mode in ['reg', 'regular']:
            print(f"{len(self.snrRatio)} SNRs * {noise_waves_len} noises, generate {clean_waves_len} noisy.")

            fullCleanWaves = clean_waves
            fullNoiseWaves = np.tile(self.noiseWaves, clean_waves_len // noise_waves_len + 1)[:clean_waves_len]
            fullSnrSequence = np.tile(self.snrRatio, clean_waves_len // snrLen + 1)[:clean_waves_len]

        elif mode in ['inc', 'increment']:
            print(f"{clean_waves_len * noise_waves_len * snrLen} noisy files "
                  f"generated by {snrLen} kinds of SNR and {noise_waves_len} noise files ")

            fullCleanWaves = np.repeat(clean_waves, noise_waves_len * snrLen)
            fullNoiseWaves = np.tile(np.repeat(self.noiseWaves, snrLen), clean_waves_len)
            fullSnrSequence = np.tile(self.snrRatio, clean_waves_len * noise_waves_len)

        elif mode in ['max', 'extra']:
            return "Please use class function 'syn_extra_mode' to run extra mode."
        else:
            return "select the right mode"

        # define output path
        fullOutPaths = []
        for cln, nos, snr in zip(fullCleanWaves, fullNoiseWaves, fullSnrSequence):
            snr = f'n{abs(snr)}' if snr < 0 else f"{snr}dB"
            fullOutPaths.append((self.noisyPath / '_'.join([cln.stem, nos.stem, snr])).with_suffix(".wav"))
        # fullOutPaths = [
        #     (self.noisyPath / '_'.join([cln.stem, nos.stem, f"{f'n{abs(snr)}' if snr < 0 else snr}dB"])).with_suffix(".wav") for
        #     cln, nos, snr in zip(fullCleanWaves, fullNoiseWaves, fullSnrSequence)]

        self.synWavesList = list(zip(fullOutPaths, fullCleanWaves, fullNoiseWaves, fullSnrSequence))

        if self.noisyPath.is_dir() and list(self.noisyPath.glob('*')):
            if is_overwrite:
                shutil.rmtree(self.noisyPath)
            else:
                warn(f"[Mixed] 路徑已存在 -> {self.noisyPath}")
                return self.synWavesList
        self.noisyPath.mkdir(parents=True, exist_ok=True)

        if add_clean:
            for cln in clean_waves:
                shutil.copy(cln, (self.noisyPath / (cln.stem + '_n0')).with_suffix(cln.suffix))

        self.__syn_pool(is_pool=is_pool, is_silence=is_silence, p_silence=p_silence)
        return self.synWavesList

    def syn_extra_mode(self, background_path='../data/backgroundnoise', background_snr=(5,)):
        # sync using background_noise
        assert not isinstance(background_snr, str), f"type error: snr_ratio can't be str"
        bgSNR = background_snr if hasattr(background_snr, '__iter__') else [background_snr]
        bgNoiseWaves, w = check_input(background_path)
        assert bgNoiseWaves, f"Error input {background_path}"
        if self.noisyPath.is_dir() and list(self.noisyPath.glob('*')):
            shutil.rmtree(self.noisyPath)
        self.noisyPath.mkdir(parents=True, exist_ok=True)
        cleanWaves = self.cleanWaves
        for cln in cleanWaves:
            shutil.copy(cln, (self.noisyPath / (cln.stem + '_n0')).with_suffix(cln.suffix))

        # 建立暫存路徑, 存放第一次混音後的檔案 + pure noise + pure back ground noise
        temporaryPath = Path('tmp_mixed')
        if temporaryPath.is_dir():
            shutil.rmtree(temporaryPath)
        temporaryPath.mkdir()

        fullCleanWaves = np.repeat(self.noiseWaves, len(bgNoiseWaves) * len(bgSNR))
        fullNoiseWaves = np.tile(np.repeat(bgNoiseWaves, len(bgSNR)), len(self.noiseWaves))
        fullSnrSequence = np.tile(bgSNR, len(self.noiseWaves) * len(bgNoiseWaves))

        fullOutPaths = [
            (temporaryPath / '_'.join(
                [cln.stem, nos.stem, f"{f'n{abs(snr)}' if snr < 0 else snr}dB"])).with_suffix(".wav") for
            cln, nos, snr in zip(fullCleanWaves, fullNoiseWaves, fullSnrSequence)]
        self.synWavesList = list(zip(fullOutPaths, fullCleanWaves, fullNoiseWaves, fullSnrSequence))
        self.__syn_pool(is_pool=False, is_silence=False)

        # increment mixed
        noiseWaves = fullOutPaths + self.noiseWaves + bgNoiseWaves
        fullCleanWaves = np.repeat(self.cleanWaves, len(noiseWaves) * len(self.snrRatio))
        fullNoiseWaves = np.tile(np.repeat(noiseWaves, len(self.snrRatio)), len(cleanWaves))
        fullSnrSequence = np.tile(self.snrRatio, len(cleanWaves) * len(noiseWaves))
        fullOutPaths = [
            (self.noisyPath / '_'.join(
                [cln.stem, nos.stem, f"{f'n{abs(snr)}' if snr < 0 else snr}dB"])).with_suffix(".wav") for
            cln, nos, snr in zip(fullCleanWaves, fullNoiseWaves, fullSnrSequence)]
        self.synWavesList = list(zip(fullOutPaths, fullCleanWaves, fullNoiseWaves, fullSnrSequence))

        self.__syn_pool(is_pool=True, is_silence=True, p_silence=0.02)
        shutil.rmtree(temporaryPath)
        return self.synWavesList

    def __syn_pool(self, is_pool: bool, is_silence: bool, p_silence=0.02):
        if is_silence and 1 > p_silence > 0:  # 將 p_silence 的noisy靜音後，只輸出 noise，作新類別 silence
            out0, cln0, nos0, snr0 = zip(*self.synWavesList)
            index = np.random.choice(range(len(self.synWavesList)), size=round(len(self.synWavesList) * p_silence))
            out, cln, nos, snr = zip(*[self.synWavesList[i] for i in index])
            snr = tuple(np.where(np.array(snr) != 0, 0, 0))
            out = tuple([o.parent / (c.stem + n.stem + '_n00.wav') for o, c, n in zip(out, cln, nos)])
            self.synWavesList = list(zip(out0 + out, cln0 + cln, nos0 + nos, snr0 + snr))

        if is_pool and len(self.synWavesList) > CoreConfig.device["pool_max"]:
            fcn = partial(Synthesizer.syn_waves, self.synWavesList, is_silence)
            with Pool(CoreConfig.device["cpu_cores"]) as pool:
                pool.map(fcn, trange(len(self.synWavesList), desc="Audio synthesis"))
        else:
            [Synthesizer.syn_waves(wave, is_silence) for wave in tqdm(self.synWavesList, desc="Audio synthesis")]

    @staticmethod
    def syn_waves(synWavesList, isSilence: bool, num=None):
        """
        synWavesList = [[save_dir, clean_file, noise_file, snr], ... ]
        """
        if CoreConfig.waveform["resample_rate"]:
            sr_clean, sr_noise = CoreConfig.waveform["resample_rate"], CoreConfig.waveform["resample_rate"]
        else:
            sr_clean, sr_noise = None, None
        save_dir, clean_file, noise_file, snr = synWavesList if num is None else synWavesList[num]
        y_clean, sr_clean = librosa.load(clean_file, sr_clean)
        y_noise, sr_noise = librosa.load(noise_file, sr_noise)
        if sr_clean != sr_noise:
            raise Exception(f'Sample rate are different. {sr_clean}, {sr_noise}')

        clean_pwr = sum(abs(y_clean) ** 2) / len(y_clean)
        if len(y_noise) < len(y_clean):
            tmp = (len(y_clean) // len(y_noise)) + 1
            y_noise = np.array([x for j in [y_noise] * tmp for x in j])
            y_noise = y_noise[:len(y_clean)]
        else:
            y_noise = y_noise[:len(y_clean)]
        y_noise = y_noise - np.mean(y_noise)
        noise_variance = clean_pwr / (10 ** (snr / 10))
        noise = np.sqrt(noise_variance) * y_noise / np.std(y_noise)

        y_noisy = noise if isSilence else y_clean + noise
        Path(save_dir).parent.mkdir(exist_ok=True, parents=True)
        wavfile.write(save_dir, sr_clean, (y_noisy * np.iinfo(np.int16).max).astype('int16'))
        # wavfile.write(save_dir, sr_clean, y_noisy.astype(np.float32))
        # librosa.output.write_wav(save_dir, y_noisy.astype(np.float32), sr_clean, norm=False)
