# -*- coding: utf-8 -*-
import platform
import re
import shutil
import librosa
import numpy as np

from copy import deepcopy
from pathlib import Path, PurePath
from scipy.io import wavfile
from multiprocessing import Pool
from functools import partial
from tqdm import tqdm, trange

from .stream import play_audio
from .tools import *
from .LoadConfig import CoreConfig

__all__ = [
    'wave_separate_by_channels',
    'save_wave',
    'load_waves',
    'concat_mfcc_',
    'concat_logspec_',
    'wav2spec',
    'spec2wav',
    'wavform2spec',
    'spec2wavform',
    'wav2mfcc',
    'check_input',
    'AudioDataManager'
]


def wave_separate_by_channels(wave, channels):
    return [wave[c::channels] for c in range(channels)]


def save_wave(file_dir, wave, sr, down_sample=None):
    Path(file_dir).parent.mkdir(exist_ok=True, parents=True)
    wave_type = wave.dtype
    if down_sample is not None:
        wave = librosa.resample(wave.astype(np.float32), sr, down_sample)
    if wave_type == np.float32:
        wave = (wave * np.iinfo(np.int16).max).astype('int16')
    wavfile.write(file_dir, sr, wave.astype('int16'))
    # librosa.output.write_wav(str(file_dir), wave.astype(np.float32), sr, norm=False)


def __wave2waveform(wave, num=None):
    info = wave if num is None else wave[num]
    if isinstance(info, (list, tuple)):
        # info = list(itertools.chain.from_iterable(convert_to_waveform(i) for i in info))
        info = [__wave2waveform(w) for w in tqdm(info, desc="Loading audio:")]
    elif isinstance(info, (str, PurePath)) and \
            Path(info).is_file() and \
            Path(info).suffix in [CoreConfig.waveform["type"]]:
        info = librosa.load(str(info), sr=None, mono=True)
    else:
        raise Exception(f"Found unexpected file {info}")
    return info


def load_waves(waves, is_pool=False):
    """
    將所有.wav 轉換為數值 (wave, sr)(type=float32, sample rate=N)
    Args:
        waves: str, list, list tree, tuple(include .txt files)
        is_pool: To use multi-processing pool
    Return:
        A list containing all audio files converted into waveform.[shape(length) is the same as input]

    result = load_waves(synWavesList, is_pool=True, cpu_croes=8)
    print(f"{len(result)}")

    """
    waves, w = check_input(waves)
    if is_pool and len(waves) > CoreConfig.device["pool_max"]:
        fcn = partial(__wave2waveform, waves)
        with Pool(CoreConfig.device["cpu_cores"]) as pool:
            result = pool.map(fcn, trange(len(waves), desc="Loading audio"))
    else:
        # result = __wave2waveform(waves[0]) if len(waves) == 1 else __wave2waveform(waves)
        result = __wave2waveform(waves)
    return result


def concat_mfcc_(wav_list, desired_samples=None):
    n_mfcc = 40
    mfcc_list = []
    for wav_dir in tqdm(wav_list):
        y, sr = librosa.load(wav_dir, sr=16000)
        if desired_samples and desired_samples < len(y):
            y = y[:desired_samples]
        mfcc = librosa.feature.mfcc(y=y, sr=sr, n_mfcc=n_mfcc).T
        mfcc = mfcc.reshape(1, mfcc.shape[0], -1)
        mfcc_list.append(mfcc)
    return np.vstack(mfcc_list)


def concat_logspec_(wav_list, sequence, forward_backward, norm, desired_samples=None):
    print("==== Concatenate all log-spectrogram from all waves ====")
    return np.vstack(
        [wav2spec(file, sequence, forward_backward, norm, desired_samples=desired_samples) for file in
         tqdm(wav_list)])


def wav2spec(file,
             sequence=False,
             forward_backward=False,
             norm=False,
             hop_length=256,
             n_fft=512,
             win_length=512,
             desired_samples=None,
             resample_rate=None):
    y, rate = librosa.load(file, sr=resample_rate, mono=True)
    if CoreConfig.waveform["resample_rate"] and rate != CoreConfig.waveform["resample_rate"]:
        warn('音檔原始取樣與預設取樣RESAMPLE_RATE值不同')

    if desired_samples and desired_samples < len(y):
        y = y[:desired_samples]
    D = librosa.stft(y,
                     n_fft=n_fft,
                     hop_length=hop_length,
                     win_length=win_length,
                     window='hamming')
    epsilon = np.finfo(float).eps
    D = D + epsilon
    Sxx = np.log10(np.abs(D) ** 2)

    if norm:
        Sxx_mean = np.mean(Sxx, axis=1).reshape(Sxx.shape[0], 1)
        # Sxx_var = np.var(Sxx, axis=1).reshape(257, 1)
        Sxx_std = np.std(Sxx, axis=1).reshape((Sxx.shape[0], 1)) + 1e-12
        Sxx_r = (Sxx - Sxx_mean) / Sxx_std  # de Std
    else:
        Sxx_r = np.array(Sxx)

    idx = 0
    # set data into 3 dim and muti-frame(frame, sample, num_frame)
    if forward_backward:
        Sxx_r = Sxx_r.T
        return_data = np.empty(
            (Sxx_r.shape[0] + 50, int(forward_backward * 2) + 1,
             int(n_fft / 2) + 1), dtype=np.float32)
        frames, dim = Sxx_r.shape
        for num, data in enumerate(Sxx_r):
            idx_start = idx - forward_backward
            idx_end = idx + forward_backward
            if idx_start < 0:
                null = np.zeros((-idx_start, dim))
                tmp = np.concatenate((null, Sxx_r[0:idx_end + 1]), axis=0)
            elif idx_end > frames - 1:
                null = np.zeros((idx_end - frames + 1, dim))
                tmp = np.concatenate((Sxx_r[idx_start:], null), axis=0)
            else:
                tmp = Sxx_r[idx_start:idx_end + 1]
            return_data[idx] = tmp
            idx += 1
        shape = return_data.shape

        if sequence:
            # return return_data[:idx]
            return return_data.reshape(1, shape[0], shape[1] * shape[2])[:, :idx]  # (1,time_step,dim)
        else:
            return return_data.reshape(shape[0], shape[1] * shape[2])[:idx]
    else:
        Sxx_r = np.array(Sxx_r).T
        shape = Sxx_r.shape
        if sequence:
            return Sxx_r.reshape(1, shape[0], shape[1])
            # return Sxx_r.reshape(shape[0], 1, shape[1])
        else:
            return Sxx_r


def spec2wav(wave_out_dir, noisy_file, enhanced_spec, squeeze=False, hop_length=256, n_fft=512, win_length=512):
    enhanced_spec = enhanced_spec.squeeze() if squeeze else enhanced_spec

    y, rate = librosa.load(noisy_file, sr=16000)
    D = librosa.stft(y.astype('float32'),
                     n_fft=n_fft,
                     hop_length=hop_length,
                     win_length=win_length,
                     window='hamming')

    epsilon = np.finfo(float).eps
    D = D + epsilon
    phase = np.exp(1j * np.angle(D))
    Sxx_r_tmp = np.array(enhanced_spec)  # enhanced log power magnitude: spec_test
    Sxx_r_tmp = np.sqrt(10 ** Sxx_r_tmp)
    Sxx_r = Sxx_r_tmp.T
    reverse = np.multiply(Sxx_r, phase)
    result = librosa.istft(reverse, hop_length=hop_length, win_length=win_length, window='hamming')

    y_out = librosa.util.fix_length(result, len(y), mode='edge')
    # y_out = y_out/np.max(np.abs(y_out))
    ## output
    Path(wave_out_dir).parent.mkdir(parents=True, exist_ok=True)
    wavfile.write(wave_out_dir, rate, (y_out * np.iinfo(np.int16).max).astype('int16'))
    # librosa.output.write_wav(str(wave_out_dir), y_out.astype(np.float32), rate, norm=False)


def wavform2spec(y,
                 sequence=False,
                 forward_backward=0,
                 norm=False,
                 hop_length=256,
                 n_fft=512,
                 win_length=512,
                 desired_samples=None):
    if desired_samples and desired_samples < len(y):
        y = y[:desired_samples]
    D = librosa.stft(y,
                     n_fft=n_fft,
                     hop_length=hop_length,
                     win_length=win_length,
                     window='hamming')
    epsilon = np.finfo(float).eps
    D = D + epsilon
    Sxx = np.log10(np.abs(D) ** 2)

    if norm:
        Sxx_mean = np.mean(Sxx, axis=1).reshape(Sxx.shape[0], 1)
        Sxx_std = np.std(Sxx, axis=1).reshape((Sxx.shape[0], 1)) + 1e-12
        Sxx_r = (Sxx - Sxx_mean) / Sxx_std
    else:
        Sxx_r = np.array(Sxx)

    idx = 0
    if forward_backward:
        Sxx_r = Sxx_r.T
        return_data = np.empty(
            (Sxx_r.shape[0] + 50, np.int32(forward_backward * 2) + 1,
             np.int32(n_fft / 2) + 1))
        frames, dim = Sxx_r.shape
        for num, data in enumerate(Sxx_r):
            idx_start = idx - forward_backward
            idx_end = idx + forward_backward
            if idx_start < 0:
                null = np.zeros((-idx_start, dim))
                tmp = np.concatenate((null, Sxx_r[0:idx_end + 1]), axis=0)
            elif idx_end > frames - 1:
                null = np.zeros((idx_end - frames + 1, dim))
                tmp = np.concatenate((Sxx_r[idx_start:], null), axis=0)
            else:
                tmp = Sxx_r[idx_start:idx_end + 1]
            return_data[idx] = tmp
            idx += 1
        shape = return_data.shape

        if sequence:
            return return_data.reshape(1, shape[0], shape[1] * shape[2])[:, :idx]
        else:
            return return_data.reshape(shape[0], shape[1] * shape[2])[:idx]
    else:
        Sxx_r = np.array(Sxx_r).T
        shape = Sxx_r.shape
        if sequence:
            return Sxx_r.reshape(1, shape[0], shape[1])
        else:
            return Sxx_r


def spec2wavform(y, enhanced_spec, sequence=False, hop_length=256, n_fft=512, win_length=512):
    if sequence:
        enhanced_spec = enhanced_spec.squeeze()
    D = librosa.stft(y.astype('float32'),
                     n_fft=n_fft,
                     hop_length=hop_length,
                     win_length=win_length,
                     window='hamming')
    epsilon = np.finfo(float).eps
    D = D + epsilon
    phase = np.exp(1j * np.angle(D))
    Sxx_r_tmp = np.array(enhanced_spec)
    Sxx_r_tmp = np.sqrt(10 ** Sxx_r_tmp)
    Sxx_r = Sxx_r_tmp.T
    reverse = np.multiply(Sxx_r, phase)
    result = librosa.istft(reverse, hop_length=hop_length, win_length=win_length, window='hamming')
    y_out = librosa.util.fix_length(result, len(y), mode='edge')
    # y_out = y_out/np.max(np.abs(y_out))
    return y_out


def wav2mfcc(y, sr):
    return librosa.feature.mfcc(y, sr, n_mfcc=40).T.reshape(1, -1, 40)


def check_input(input):
    """
    檢查/尋找 副檔名為 WAVE_TYPE 的檔案
    Args:
        input : anything. file path, list of file path, .txt file contains file path, list of .txt files, ...
    Return:
        list1, list2 (list1 if file with suffix "WAVE_TYPE" else list2)
    """
    rightList = []
    wrongList = []
    if isinstance(input, (list, tuple)):
        for item in input:
            r, w = check_input(item)
            rightList.extend(r)
            wrongList.extend(w)
    elif input and isinstance(input, (str, PurePath)):
        input = Path(input)
        if input.is_file():
            if input.suffix == '.txt':
                with open(input, 'r') as f:
                    input = [x.strip() for x in f.read().strip().splitlines() if len(x.strip())]
                    for item in input:
                        r, w = check_input(item)
                        rightList.extend(r)
                        wrongList.extend(w)
            elif input.suffix == CoreConfig.waveform["type"]:
                rightList.append(input)
            else:
                wrongList.append(str(input))
        elif input.is_dir():
            rightList.extend(list(input.rglob(f"*{CoreConfig.waveform['type']}")))
        else:
            wrongList.append(str(input))
    else:
        wrongList.append(input)
    return rightList, wrongList


class AudioDataManager:
    def __init__(self):
        self.suffix = '.wav'
        self.rate_ = None
        self.ch_ = 1
        self.default_dict = {"dir": None, "wave": [], "sr": None, "ch": self.ch_}
        self.reset_defaults()

    def reset_defaults(self):
        self.audios_ = dict()

    reset = reset_defaults

    def set_params(self, sr: int, ch: int):
        assert sr > 0, "sr setting error"
        self.ch_ = ch if 3 > ch > 0 else 1
        self.rate_ = sr
        self.default_dict = {"dir": None, "wave": [] * self.ch_, "sr": sr, "ch": self.ch_}

    def __call__(self, wave_name):
        return self.get_item(wave_name)

    def __getitem__(self, wave_name):
        if isinstance(wave_name, str):
            return self.get_item(wave_name)
        elif isinstance(wave_name, int):
            return self.get_item(list(self.get_names())[wave_name])

    def get_item(self, name):
        try:
            return self.audios_[name]
        except Exception as e:
            print(f"[AudioManager] {e}")
            return None

    def __len__(self):
        return len(self.audios_)

    def get_wave(self, name):
        item = self.__getitem__(name)
        if item is not None:
            return item['wave']
        else:
            return None

    def get_sample(self, name):
        item = self.__getitem__(name)
        if item is not None:
            return item['sr']
        else:
            return None

    def add_local_file(self, dir, name: str = None):
        dir = Path(dir)
        assert dir.is_file(), f'[AudioManager] (add_local_file) File not exists ->{dir}'
        name = dir.stem if name is None else name
        if name in self.get_names():
            warn("[AudioManager] (add_local_file) Name exist")
            return None
        self.audios_[name] = deepcopy(self.default_dict)
        self.audios_[name]["dir"] = str(dir)
        wave_, rate_ = load_waves(dir)[0]
        self.audios_[name]["wave"].append(wave_)
        self.audios_[name]["sr"] = rate_
        self.audios_[name]["ch"] = 1
        # return wave_, rate_

    def add_new_file(self, name: str = None, dir=None, wave=None, sr: int = None, ch: int = None):

        if name is None:
            if dir is None:
                return warn('[AudioManager] (add_new_file) Please set a name or dir')
            else:
                name = Path(dir).stem
        if name in self.get_names():
            warn("Name exist")
        else:
            sr = self.rate_ if sr is None else sr
            ch = self.ch_ if ch is None else ch
            self.audios_[name] = deepcopy(self.default_dict)
            if dir is not None:
                self.audios_[name]["dir"] = str(dir)
            if wave is not None:
                if len(wave) == 1 and ch > 0:
                    wave = wave_separate_by_channels(wave[0], ch)
                    for i in range(ch):
                        self.audios_[name]["wave"].append(wave[i])
                else:
                    self.del_element(name)
                    return warn('[AudioManager] (add_new_file) Unexpected wave or ch')
            self.audios_[name]["sr"] = sr
            self.audios_[name]["ch"] = ch

    def add_name(self, name: str):
        self.audios_[name] = self.default_dict.copy()

    def add_dir(self, name: str, dir: str):
        self.audios_[name]["dir"] = dir

    def add_wave(self, name: str, wave):
        self.audios_[name]["wave"].append(wave)

    def add_sr(self, name, sr):
        self.audios_[name]["sr"] = sr

    def add_ch(self, name, ch):
        self.audios_[name]["ch"] = ch

    def del_element(self, name):
        try:
            del self.audios_[name]
        except Exception as e:
            print(e)

    def get_names(self):
        return self.audios_.keys()

    def get_values(self):
        return self.audios_.values()

    def save_(self, name, dst_path=None, down_sample=None):
        if self.__len__() > 0:
            item = self.get_item(name)
            if item is None:
                print("[AudioManager] (save_) Data not found!")
                return
            else:
                dir_ = item['dir'] if dst_path is None else (Path(dst_path) / name).with_suffix(self.suffix)
                if self.ch_ > 1:
                    pth = Path(dir_).parent
                    for i, w in enumerate(item['wave']):
                        d = (pth / f"{name}_ch{i}").with_suffix(self.suffix)
                        save_wave(d, w, item['sr'], down_sample=down_sample)
                else:
                    save_wave(dir_, item['wave'][0], item['sr'], down_sample=down_sample)
        else:
            print("[AudioManager] (save_) Data empty!")

    def save_all(self, dst_path=None, down_sample=None):
        if self.__len__() > 0:
            for n in self.get_names():
                item = self.get_item(n)
                if item is None:
                    print("[AudioManager] (save_all) Data not found!")
                    return False
                else:
                    dir_ = item['dir'] if dst_path is None else (Path(dst_path) / n).with_suffix(self.suffix)
                    if self.ch_ is not None and self.ch_ > 1:
                        pth = Path(dir_).parent
                        for i, w in enumerate(item['wave']):
                            d = (pth / f"{n}_ch{i}").with_suffix(self.suffix)
                            save_wave(d, w, item['sr'], down_sample=down_sample)
                    else:
                        save_wave(dir_, item['wave'][0], item['sr'], down_sample=down_sample)
            print("[AudioManager] Save all wave success")
            return True
        else:
            print("[AudioManager] (save_all) Data empty!")
            return False

    def play_audio(self, name, volume: float):
        item = self.__getitem__(name)
        if item is not None:
            w = item['wave'][0]
            dtype_ = w.dtype
            play_audio(wav=(w * volume).astype(dtype_), rate=item['sr'])
        else:
            print(f"[AudioManager] Play audio error, wave:{name}")

# if __name__ == '__main__':
#     # To Check list or find
#     wave_list, other = check_input('../data/train/clean/big')
#
#     # ma = AudioManager()
#     # ma.add("abc", "./asd.wav")
#     # ma['abc']
#     # wavfile.read()
#
#     # # combiner('a', 'b', exec='yes')
#     # ## set input path
#     input = '../data/train/clean/big'
#     # # input = ['data/clean', 'data/noise']
#     # # input = 'input.txt'
#     ss = Synthesizer(clean='../data/train/clean/big', noise='../data/train/noise')
#
#     # cln_waves, nos_waves = ss.wave_input()  # get wave list
#     #
#     cln_waveform, nos_waveform = ss.waveform(is_pool=False)  # load wave
#
#     ss.syn('reg')  # synthesize wave
#     #
#     # ## ==========================================
#     # ## get waveform from wave list
#     waves1 = load_waves(cln_waves, is_pool=True)
#     waves2 = load_waves(cln_waves, is_pool=False)
#     ## get waveform from a wave
#     wave = load_waves(cln_waves[0], is_pool=False)
